"""Download data needed for the examples"""

from __future__ import annotations

if __name__ == "__main__":

    from os import path, makedirs, remove
    from download_sample_data import bar as progressbar

    import pandas as pd
    import numpy as np
    import sys

    try:
        import requests
    except ImportError:
        print('Download script required requests package: conda install requests')
        sys.exit(1)

    def _download_dataset(url):
        r = requests.get(url, stream=True)
        output_path = path.split(url)[1]
        with open(output_path, 'wb') as f:
            total_length = int(r.headers.get('content-length'))
            for chunk in progressbar(r.iter_content(chunk_size=1024), expected_size=(total_length/1024) + 1):
                if chunk:
                    f.write(chunk)
                    f.flush()

    examples_dir = path.dirname(path.realpath(__file__))
    data_dir = path.join(examples_dir, 'data')
    if not path.exists(data_dir):
        makedirs(data_dir)

    # Taxi data
    def latlng_to_meters(df, lat_name, lng_name):
        lat = df[lat_name]
        lng = df[lng_name]
        origin_shift = 2 * np.pi * 6378137 / 2.0
        mx = lng * origin_shift / 180.0
        my = np.log(np.tan((90 + lat) * np.pi / 360.0)) / (np.pi / 180.0)
        my = my * origin_shift / 180.0
        df.loc[:, lng_name] = mx
        df.loc[:, lat_name] = my

    taxi_path = path.join(data_dir, 'nyc_taxi.csv')
    if not path.exists(taxi_path):
        print("Downloading Taxi Data...")
        url = ('https://storage.googleapis.com/tlc-trip-data/2015/'
               'yellow_tripdata_2015-01.csv')

        _download_dataset(url)
        df = pd.read_csv('yellow_tripdata_2015-01.csv')

        print('Filtering Taxi Data')
        df = df.loc[(df.pickup_longitude < -73.75) &
                    (df.pickup_longitude > -74.15) &
                    (df.dropoff_longitude < -73.75) &
                    (df.dropoff_longitude > -74.15) &
                    (df.pickup_latitude > 40.68) &
                    (df.pickup_latitude < 40.84) &
                    (df.dropoff_latitude > 40.68) &
                    (df.dropoff_latitude < 40.84)].copy()

        print('Reprojecting Taxi Data')
        latlng_to_meters(df, 'pickup_latitude', 'pickup_longitude')
        latlng_to_meters(df, 'dropoff_latitude', 'dropoff_longitude')
        df.rename(columns={'pickup_longitude': 'pickup_x', 'dropoff_longitude': 'dropoff_x',
                           'pickup_latitude': 'pickup_y', 'dropoff_latitude': 'dropoff_y'},
                  inplace=True)
        df.to_csv(taxi_path, index=False)
        remove('yellow_tripdata_2015-01.csv')


    print("\nAll data downloaded.")



"""Error handling"""

syntax_error = """

  File {filename}, line {lineno}
    {line}
    {pointer}

{error}: {msg}
"""

class DataShapeSyntaxError(SyntaxError):
    """
    Makes datashape parse errors look like Python SyntaxError.
    """
    def __init__(self, lexpos, filename, text, msg=None):
        self.lexpos = lexpos
        self.filename = filename
        self.text = text
        self.msg = msg or 'invalid syntax'
        self.lineno = text.count('\n', 0, lexpos) + 1
        # Get the extent of the line with the error
        linestart = text.rfind('\n', 0, lexpos)
        if linestart < 0:
            linestart = 0
        else:
            linestart = linestart + 1
        lineend = text.find('\n', lexpos)
        if lineend < 0:
            lineend = len(text)
        self.line = text[linestart:lineend]
        self.col_offset = lexpos - linestart

    def __str__(self):
        pointer = ' ' * self.col_offset + '^'

        return syntax_error.format(
            filename=self.filename,
            lineno=self.lineno,
            line=self.line,
            pointer=pointer,
            msg=self.msg,
            error=self.__class__.__name__,
        )

    def __repr__(self):
        return str(self)



"""
Binary graphical composition operators

See https://www.cairographics.org/operators/; more could easily be added from there.
"""

from __future__ import annotations

import numba as nb
import numpy as np
import os

image_operators = ('over', 'add', 'saturate', 'source')
array_operators = ('add_arr', 'max_arr', 'min_arr', 'source_arr')
__all__ = ('composite_op_lookup', 'validate_operator') + image_operators + array_operators


def validate_operator(how, is_image):
    name = how if is_image else how + '_arr'
    if is_image:
        if name not in image_operators:
            raise ValueError('Operator %r not one of the supported image operators: %s'
                            % (how, ', '.join(repr(el) for el in image_operators)))
    elif name not in array_operators:
        raise ValueError('Operator %r not one of the supported array operators: %s'
                        % (how, ', '.join(repr(el[:-4]) for el in array_operators)))


@nb.jit('(uint32,)', nopython=True, nogil=True, cache=True)
def extract_scaled(x):
    """Extract components as float64 values in [0.0, 1.0]"""
    r = np.float64(( x        & 255) / 255)
    g = np.float64(((x >>  8) & 255) / 255)
    b = np.float64(((x >> 16) & 255) / 255)
    a = np.float64(((x >> 24) & 255) / 255)
    return r, g, b, a


@nb.jit('(float64, float64, float64, float64)', nopython=True,
        nogil=True, cache=True)
def combine_scaled(r, g, b, a):
    """Combine components in [0, 1] to rgba uint32"""
    r2 = min(255, np.uint32(r * 255))
    g2 = min(255, np.uint32(g * 255))
    b2 = min(255, np.uint32(b * 255))
    a2 = min(255, np.uint32(a * 255))
    return np.uint32((a2 << 24) | (b2 << 16) | (g2 << 8) | r2)


jit_enabled = os.environ.get('NUMBA_DISABLE_JIT', '0') == '0'


if jit_enabled:
    extract_scaled.disable_compile()
    combine_scaled.disable_compile()

# Lookup table for storing compositing operators by function name
composite_op_lookup = {}


def operator(f):
    """Define and register a new image composite operator"""

    if jit_enabled:
        f2 = nb.vectorize(f)
        f2._compile_for_argtys((nb.types.uint32, nb.types.uint32))
        f2._frozen = True
    else:
        f2 = np.vectorize(f)

    composite_op_lookup[f.__name__] = f2
    return f2


@operator
def source(src, dst):
    if src & 0xff000000:
        return src
    else:
        return dst


@operator
def over(src, dst):
    sr, sg, sb, sa = extract_scaled(src)
    dr, dg, db, da = extract_scaled(dst)

    factor = 1 - sa
    a = sa + da * factor
    if a == 0:
        return np.uint32(0)
    r = (sr * sa + dr * da * factor)/a
    g = (sg * sa + dg * da * factor)/a
    b = (sb * sa + db * da * factor)/a
    return combine_scaled(r, g, b, a)


@operator
def add(src, dst):
    sr, sg, sb, sa = extract_scaled(src)
    dr, dg, db, da = extract_scaled(dst)

    a = min(1, sa + da)
    if a == 0:
        return np.uint32(0)
    r = (sr * sa + dr * da)/a
    g = (sg * sa + dg * da)/a
    b = (sb * sa + db * da)/a
    return combine_scaled(r, g, b, a)


@operator
def saturate(src, dst):
    sr, sg, sb, sa = extract_scaled(src)
    dr, dg, db, da = extract_scaled(dst)

    a = min(1, sa + da)
    if a == 0:
        return np.uint32(0)
    factor = min(sa, 1 - da)
    r = (factor * sr + dr * da)/a
    g = (factor * sg + dg * da)/a
    b = (factor * sb + db * da)/a
    return combine_scaled(r, g, b, a)



def arr_operator(f):
    """Define and register a new array composite operator"""

    if jit_enabled:
        f2 = nb.vectorize(f)
        f2._compile_for_argtys(
           (nb.types.int32, nb.types.int32))
        f2._compile_for_argtys(
           (nb.types.int64, nb.types.int64))
        f2._compile_for_argtys(
            (nb.types.float32, nb.types.float32))
        f2._compile_for_argtys(
            (nb.types.float64, nb.types.float64))
        f2._frozen = True
    else:
        f2 = np.vectorize(f)

    composite_op_lookup[f.__name__] = f2
    return f2


@arr_operator
def source_arr(src, dst):
    if src:
        return src
    else:
        return dst

@arr_operator
def add_arr(src, dst):
    return src + dst

@arr_operator
def max_arr(src, dst):
    return max([src,  dst])

@arr_operator
def min_arr(src, dst):
    return min([src,  dst])



from __future__ import annotations
from datashader.glyphs.line import LinesXarrayCommonX
from datashader.glyphs.quadmesh import _QuadMeshLike
from datashader.data_libraries.pandas import default
from datashader.core import bypixel
import xarray as xr
from datashader.utils import Dispatcher


try:
    import cupy
except Exception:
    cupy = None

glyph_dispatch = Dispatcher()


@bypixel.pipeline.register(xr.Dataset)
def xarray_pipeline(xr_ds, schema, canvas, glyph, summary, *, antialias=False):
    cuda = False
    if cupy:
        if isinstance(glyph, LinesXarrayCommonX):
            cuda = isinstance(xr_ds[glyph.y].data, cupy.ndarray)
        else:
            cuda = isinstance(xr_ds[glyph.name].data, cupy.ndarray)

    if not xr_ds.chunks:
        return glyph_dispatch(
            glyph, xr_ds, schema, canvas, summary, antialias=antialias, cuda=cuda)
    else:
        from datashader.data_libraries.dask_xarray import dask_xarray_pipeline
        return dask_xarray_pipeline(
            glyph, xr_ds, schema, canvas, summary, antialias=antialias, cuda=cuda)


# Default to default pandas implementation
glyph_dispatch.register(_QuadMeshLike)(default)
glyph_dispatch.register(LinesXarrayCommonX)(default)



from abc import ABC

from ..coretypes import (
    DataShape,
    DateTime,
    Function,
    Option,
    Record,
    String,
    Time,
    TimeDelta,
    Tuple,
    Units,
)
from ..dispatch import dispatch


def _fmt_path(path):
    """Format the path for final display.

    Parameters
    ----------
    path : iterable of str
        The path to the values that are not equal.

    Returns
    -------
    fmtd : str
        The formatted path to put into the error message.
    """
    if not path:
        return ''
    return 'path: _' + ''.join(path)


@dispatch(DataShape, DataShape)
def assert_dshape_equal(a, b, check_dim=True, path=None, **kwargs):
    """Assert that two dshapes are equal, providing an informative error
    message when they are not equal.

    Parameters
    ----------
    a, b : dshape
        The dshapes to check for equality.
    check_dim : bool, optional
        Check shapes for equality with respect to their dimensions.
        default: True
    check_tz : bool, optional
        Checks times and datetimes for equality with respect to timezones.
        default: True
    check_timedelta_unit : bool, optional
        Checks timedeltas for equality with respect to their unit (us, ns, ...).
        default: True
    check_str_encoding : bool, optional
        Checks strings for equality with respect to their encoding.
        default: True
    check_str_fixlen : bool, optional
        Checks string for equality with respect to their fixlen.
        default: True
    check_record_order : bool, optional
        Checks records for equality with respect to the order of the fields.
        default: True

    Raises
    ------
    AssertionError
        Raised when the two dshapes are not equal.
    """
    ashape = a.shape
    bshape = b.shape

    if path is None:
        path = ()

    if check_dim:
        for n, (adim, bdim) in enumerate(zip(ashape, bshape)):
            if adim != bdim:
                path += '.shape[%d]' % n,
                raise AssertionError(
                    'dimensions do not match: %s != %s%s\n%s' % (
                        adim,
                        bdim,
                        ('\n%s != %s' % (
                            ' * '.join(map(str, ashape)),
                            ' * '.join(map(str, bshape)),
                        )) if len(a.shape) > 1 else '',
                        _fmt_path(path),
                    ),
                )

    path += '.measure',
    assert_dshape_equal(
        a.measure,
        b.measure,
        check_dim=check_dim,
        path=path,
        **kwargs
    )


class Slotted(ABC):
    @classmethod
    def __subclasshook__(cls, subcls):
        return hasattr(subcls, '__slots__')


@assert_dshape_equal.register(Slotted, Slotted)
def _check_slots(a, b, path=None, **kwargs):
    if type(a) != type(b):
        return _base_case(a, b, path=path, **kwargs)

    assert a.__slots__ == b.__slots__, 'slots mismatch: %r != %r\n%s' % (
        a.__slots__, b.__slots__, _fmt_path(path),
    )
    if path is None:
        path = ()
    for slot in a.__slots__:
        assert getattr(a, slot) == getattr(b, slot), \
            "%s %ss do not match: %r != %r\n%s" % (
                type(a).__name__.lower(),
                slot,
                getattr(a, slot),
                getattr(b, slot),
                _fmt_path(path + ('.' + slot,)),
            )


@assert_dshape_equal.register(object, object)
def _base_case(a, b, path=None, **kwargs):
    assert a == b, '%s != %s\n%s' % (a, b, _fmt_path(path))


@dispatch((DateTime, Time), (DateTime, Time))
def assert_dshape_equal(a, b, path=None, check_tz=True, **kwargs):
    if type(a) != type(b):
        return _base_case(a, b)
    if check_tz:
        _check_slots(a, b, path)


@dispatch(TimeDelta, TimeDelta)
def assert_dshape_equal(a, b, path=None, check_timedelta_unit=True, **kwargs):
    if check_timedelta_unit:
        _check_slots(a, b, path)


@dispatch(Units, Units)
def assert_dshape_equal(a, b, path=None, **kwargs):
    if path is None:
        path = ()

    assert a.unit == b.unit, '%s units do not match: %r != %s\n%s' % (
        type(a).__name__.lower(), a.unit, b.unit, _fmt_path(path + ('.unit',)),
    )

    path.append('.tp')
    assert_dshape_equal(a.tp, b.tp, **kwargs)


@dispatch(String, String)
def assert_dshape_equal(a,
                        b,
                        path=None,
                        check_str_encoding=True,
                        check_str_fixlen=True,
                        **kwargs):
    if path is None:
        path = ()
    if check_str_encoding:
        assert a.encoding == b.encoding, \
            'string encodings do not match: %r != %r\n%s' % (
                a.encoding, b.encoding, _fmt_path(path + ('.encoding',)),
            )

    if check_str_fixlen:
        assert a.fixlen == b.fixlen, \
            'string fixlens do not match: %d != %d\n%s' % (
                a.fixlen, b.fixlen, _fmt_path(path + ('.fixlen',)),
            )


@dispatch(Option, Option)
def assert_dshape_equal(a, b, path=None, **kwargs):
    if path is None:
        path = ()
    path += '.ty',
    return assert_dshape_equal(a.ty, b.ty, path=path, **kwargs)


@dispatch(Record, Record)
def assert_dshape_equal(a, b, check_record_order=True, path=None, **kwargs):
    afields = a.fields
    bfields = b.fields

    assert len(afields) == len(bfields), \
        'records have mismatched field counts: %d != %d\n%r != %r\n%s' % (
            len(afields), len(bfields), a.names, b.names, _fmt_path(path),
        )

    if not check_record_order:
        afields = sorted(afields)
        bfields = sorted(bfields)

    if path is None:
        path = ()
    for n, ((aname, afield), (bname, bfield)) in enumerate(
            zip(afields, bfields)):

        assert aname == bname, \
            'record field name at position %d does not match: %r != %r\n%s' % (
                n, aname, bname, _fmt_path(path),
            )

        assert_dshape_equal(
            afield,
            bfield,
            path=path + ('[%s]' % repr(aname),),
            check_record_order=check_record_order,
            **kwargs
        )


@dispatch(Tuple, Tuple)
def assert_dshape_equal(a, b, path=None, **kwargs):
    assert len(a.dshapes) == len(b.dshapes), \
        'tuples have mismatched field counts: %d != %d\n%r != %r\n%s' % (
            len(a.dshapes), len(b.dshapes), a, b, _fmt_path(path),
        )

    if path is None:
        path = ()
    path += '.dshapes',
    for n, (ashape, bshape) in enumerate(zip(a.dshapes, b.dshapes)):
        assert_dshape_equal(
            ashape,
            bshape,
            path=path + ('[%d]' % n,),
            **kwargs
        )


@dispatch(Function, Function)
def assert_dshape_equal(a, b, path=None, **kwargs):
    assert len(a.argtypes) == len(b.argtypes),\
        'functions have different arities: %d != %d\n%r != %r\n%s' % (
            len(a.argtypes), len(b.argtypes), a, b, _fmt_path(path),
        )

    if path is None:
        path = ()
    for n, (aarg, barg) in enumerate(zip(a.argtypes, b.argtypes)):
        assert_dshape_equal(
            aarg,
            barg,
            path=path + ('.argtypes[%d]' % n,), **kwargs
        )
    assert_dshape_equal(
        a.restype,
        b.restype,
        path=path + ('.restype',),
        **kwargs
    )



from __future__ import annotations
from enum import Enum
from typing import NamedTuple, TYPE_CHECKING


# Enum used to specify how the second stage aggregation is performed
# for 2-stage antialiased lines.
class AntialiasCombination(Enum):
    SUM_1AGG = 1
    SUM_2AGG = 2
    MIN = 3
    MAX = 4
    FIRST = 5
    LAST = 6


class AntialiasStage2(NamedTuple):
    """Configuration for second-stage combination of a single antialiased reduction."""
    combination: AntialiasCombination
    zero: float
    n_reduction: bool = False
    categorical: bool = False


if TYPE_CHECKING:
    UnzippedAntialiasStage2 = \
        tuple[tuple[AntialiasCombination], tuple[float], tuple[bool], tuple[bool]]


def two_stage_agg(antialias_stage_2: UnzippedAntialiasStage2 | None):
    """Information used to perform the correct stage 2 aggregation."""
    if not antialias_stage_2:
        # Not using antialiased lines, doesn't matter what is returned.
        return False, False

    aa_combinations = antialias_stage_2[0]

    # A single combination in (SUM_2AGG, FIRST, LAST, MIN) means that a 2-stage
    # aggregation will be used, otherwise use a 1-stage aggregation that is
    # faster.
    use_2_stage_agg = False
    for comb in aa_combinations:
        if comb in (AntialiasCombination.SUM_2AGG, AntialiasCombination.MIN,
                    AntialiasCombination.FIRST, AntialiasCombination.LAST):
            use_2_stage_agg = True
            break

    # Boolean overwrite flag is used in _full_antialias() is True to overwrite
    # pixel values (using max of previous and new values) or False for the more
    # complicated correction algorithm. Prefer overwrite=True for speed, but
    # any SUM_1AGG implies overwrite=False.
    overwrite = True
    for comb in aa_combinations:
        if comb == AntialiasCombination.SUM_1AGG:
            overwrite = False
            break

    return overwrite, use_2_stage_agg



"""Tests for antialiased line drawing in Datashader"""

# The single-pixel width tests consist of 6 test cases, each drawing a
# different set of lines on a square canvas. Tests 1 to 4 are all sets of lines
# that start from the same point of origin and then disperse across the canvas
# to fill either an upper or a lower triangle. What follows is a simple
# ascii-art based cartoon to illustrate the setup. It only has 3 lines instead
# of nine and it is a cartoon within the limits of being able to draw angled
# lines on a square grid. The origin is the lower left corner and all nine
# lines are in the triangle below the diagonal.
#
# +---------------------------------------------+
# |                                             |
# |                                         *   |
# |                                       *     |
# |                                     *       |
# |                                   *         |
# |                                 *           |
# |                               *             |
# |                             *               |
# |                           *                 |
# |                         *                   |
# |                       *                     |
# |                     *                    *  |
# |                   *                  *      |
# |                 *                *          |
# |               *              *              |
# |             *            *                  |
# |           *          *                      |
# |         *        *                          |
# |       *      *                              |
# |     *    *                                  |
# |   *  *                                      |
# | ******************************************* |
# |                                             |
# +---------------------------------------------+
#
# This is test_001. The following three tests 002, 003 and 004 are:
# * test_002: lower left corner as origin, lines in upper triangle
# * test_003: upper right corner as origin, lines in upper triangle
# * test_004: upper right corner as origin, lines in lower triangle
#
# The tests 005 and 006 use a different pattern, which is a combination of
# vertical and diagonal lines, where the angle of the diagonal lines
# 'increases'. Here is another cartoon to illustrate these tests cases that
# shows three verticals and two diagonals. As you can see, the angle between
# the vertical and the following diagonal increases from left to right.

# +---------------------------------------------+
# | *           *                    *          |
# | **          **                   *          |
# | * *         * *                  *          |
# | * *         *  *                 *          |
# | *  *        *   *                *          |
# | *  *        *    *               *          |
# | *   *       *     *              *          |
# | *   *       *      *             *          |
# | *    *      *       *            *          |
# | *    *      *        *           *          |
# | *     *     *         *          *          |
# | *     *     *          *         *          |
# | *      *    *           *        *          |
# | *      *    *            *       *          |
# | *       *   *             *      *          |
# | *       *   *              *     *          |
# | *        *  *               *    *          |
# | *        *  *                *   *          |
# | *         * *                 *  *          |
# | *         * *                  * *          |
# | *          **                   **          |
# | *           *                    *          |
# +---------------------------------------------+
#
# The big difference between 005 and 006 is that in 005 each line is individual
# line whereas for 006 it is a multi-segment line, and each vertex is listed
# only a single time. Datashader then "connects the dots" as it were.
#
# Test 007 tests the edge case, where we draw an almost straight line between
# corners with only a single pixel offset. This is to ensure that anti-aliasing
# does not try to draw pixels that  are out of bounds. Importantly, this needs
# to be run with Numba disabled, since Numba does not do OOB checking by
# default.
#
# +---------------------------------------------+
# | *                     **********************|
# |***********************                     *|
# | *                                          *|
# | *                                          *|
# | *                                          *|
# | *                                          *|
# | *                                          *|
# | *                                          *|
# | *                                          *|
# | *                                          *|
# | *                                          *|
# |*                                          * |
# |*                                          * |
# |*                                          * |
# |*                                          * |
# |*                                          * |
# |*                                          * |
# |*                                          * |
# |*                                          * |
# |*                                          * |
# |*                       *********************|
# |************************                   * |
# +---------------------------------------------+
#
#
# So for each of these 7 patterns, we test a regular and a clipped version
# (the canvas is clipped to a region in the center) in both the normal and the
# anti-aliased drawing mode. This ensures that lines can be drawn in all
# directions and that clipping a canvas works. Tests 005 and 006 ensure that
# multi-line segments can be drawn correctly too.
#
# Each tests case comes with an image and a saved xarray file (serialized in
# the NetCDF format). The image is to aid visual inspection of the algorithm
# and the quality of the anti-aliasing. The serialized xarray is for automated
# testing. The __name__ == '__main__'  section can be used to re-generate both
# the NetCDF files and the PNG files.


import os

import xarray as xr
import datashader as ds
import pandas as pd
import numpy as np
import pytest


cm = pytest.importorskip('matplotlib.cm')
binary = cm.binary

# Data directory for saving and loading test data
datadir = os.path.join(os.path.dirname(__file__), 'data')

# The colormap used for rendering
cmap01 =([tuple(v*255. for v in tuple(rgb)[:3])
          for rgb in binary(np.linspace(0, 1))])

# The normal sized canvas
regular_cvs = ds.Canvas(plot_width=50, plot_height=50,
                        x_range=(0, 49), y_range=(0, 49))

# The reduced or clipped canvas
reduced_cvs = ds.Canvas(plot_width=20, plot_height=20,
                        x_range=(14, 34), y_range=(14, 34))

# Datashader options for anti-alias and canvas size
antialias_options = ((True, "antialias"), (False, "noaa"))
canvas_options = ((regular_cvs, "normal"), (reduced_cvs, "clipped"))


def draw_line(cvs, p1, p2, antialias):
    """Draw a single line.

    Parameters
    ----------
    cvs: canvas
      A Datashader canvas
    p1: tuple
      The first vertex of the line
    p2: tuple
      The second vertex of the line
    antialias: boolean
      To anti-alias or not is the question

    Returns
    -------
    agg: A Datashader aggregator (xarray)

    """
    xs, ys = np.array([p1[0], p2[0]]), np.array([p1[1], p2[1]])
    points = pd.DataFrame({'x': xs, 'y': ys, 'val': 5.0})
    return cvs.line(points, 'x', 'y', agg=ds.reductions.max("val"),
                    antialias=antialias)


def draw_lines(cvs, points, antialias):
    """Draw multiple line.

    Parameters
    ----------
    cvs: canvas
      A Datashader canvas
    points: list of tuple of tuple
      The lines to render as a list of tuples, where each tuple represents a
      line consisting of two tuples each containing two scalars describing the
      two vertices of the line.
    antialias: boolean
      To anti-alias or not is the question

    Returns
    -------
    agg: A Datashader aggregator (xarray)
    """
    aggs = []
    for ((x1, y1), (x2, y2)) in points:
        aggs.append(draw_line(cvs, (x1, y1), (x2, y2), antialias))
    return xr.concat(aggs, 'stack').sum(dim='stack')


def draw_multi_segment_line(cvs, points, antialias):
    """Draw multi-line segment line.

    Parameters
    ----------
    cvs: canvas
      A Datashader canvas
    points: list of tuples
      List of tuples of two scalars that represent each of the vertices in the
      multi-segment line.
    antialias: boolean
      To anti-alias or not is the question

    Returns
    -------
    agg: A Datashader aggregator (xarray)
    """
    x, y = [], []
    for (x1, y1) in points:
        x.append(x1)
        y.append(y1)
    xs, ys = np.array(x), np.array(y)
    points = pd.DataFrame({'x': xs, 'y': ys, 'val': 5.0})
    agg = cvs.line(points, 'x', 'y', agg=ds.reductions.max("val"),
                    antialias=antialias)
    # This is required for the line to render properly
    return xr.concat([agg], 'stack').sum(dim='stack')


def shade(aggregators, cmap=cmap01):
    """Shade/render the aggregator.

    Parameters
    ----------
    aggregators: xarray
      The aggregator(s)  to shade
    cmap: color map
      The colormap to use

    Returns
    -------
    img: xarray
      The shaded image.
    """
    img = ds.transfer_functions.shade(aggregators, cmap=cmap)
    img = ds.transfer_functions.set_background(img, '#ffffff')
    return img

def save_to_image(img, filename):
    """Save a shaded image as PNG file.

    Parameters
    ----------
    img: xarray
      The image to save
    filename: unicode
      The name of the file to save to, 'png' extension will be appended.

    """
    filename = os.path.join(datadir, filename + '.png')
    print('Saving: ' + filename)
    img.to_pil().save(filename)

def save_to_netcdf(img, filename):
    """Save a shaded image as NetCDF file.

    Parameters
    ----------
    img: xarray
      The image to save
    filename: unicode
      The name of the file to save to, 'nc' extension will be appended.

    """
    filename = os.path.join(datadir, filename + '.nc')
    print('Saving: ' + filename)
    img.to_netcdf(filename)


def load_from_netcdf(filename):
    """Load a shaded image from NetCDF file.

    Parameters
    ----------
    filename: unicode
      The name of the file to load from.

    Returns
    -------
    img: xarray
      The loaded image.

    """
    filename = os.path.join(datadir, filename + '.nc')
    return xr.open_dataarray(filename)

# All test generators return a 'points' list and the name of the point set

def generate_test_001():
    points = []
    for a in range(1, 55, 6):
        points.append(((1, 1), (49, a)))
    return points, "test_001"

def generate_test_002():
    points = []
    for a in range(1, 55, 6):
        points.append(((1, 1), (a, 49)))
    return points, "test_002"

def generate_test_003():
    points = []
    for a in range(1, 55, 6):
        points.append(((49, 49), (1, a)))
    return points, "test_003"

def generate_test_004():
    points = []
    for a in range(1, 55, 6):
        points.append(((49, 49), (a, 1)))
    return points, "test_004"

def generate_test_005():
    points = [
        ((1, 1),  (1, 49)),
        ((1, 49), (2, 1)),
        ((2, 1),  (2, 49)),
        ((2, 49), (4, 1)),
        ((4, 1),  (4, 49)),
        ((4, 49), (8, 1)),
        ((8, 1),  (8, 49)),
        ((8, 49), (16, 1)),
        ((16, 1), (16, 49)),
        ((16, 49),(32, 1)),
        ((32, 1), (32, 49)),
        ((32, 49),(49, 1)),
        ((49, 1), (49, 49)),
    ]
    return points, "test_005"

def generate_test_006():
    points = [
        (1, 1),
        (1, 49),
        (2, 1),
        (2, 49),
        (4, 1),
        (4, 49),
        (8, 1),
        (8, 49),
        (16, 1),
        (16, 49),
        (32, 1),
        (32, 49),
        (49, 1),
        (49, 49),
    ]
    return points, "test_006"

def generate_test_007():
    points = [
        ((0.5, 0.5),  (1.5, 48.5)),
        ((0.5, 0.5),  (48.5, 1.5)),
        ((48.5, 48.5),  (47.5, 0.5)),
        ((48.5, 48.5),  (0.5, 47.5)),
    ]
    return points, "test_007"

def generate_test_images():
    """Generate all test images.

    Returns
    -------
    results: dict
      A dictionary mapping test case name to xarray images.
    """
    results = {}
    for antialias, aa_descriptor  in antialias_options:
        for canvas, canvas_descriptor in canvas_options:
            for func in (generate_test_001,
                         generate_test_002,
                         generate_test_003,
                         generate_test_004,
                         generate_test_005,
                         generate_test_007,
                        ):
                points, name = func()
                aggregators = draw_lines(canvas, points, antialias)
                img = shade(aggregators, cmap=cmap01)
                description = "{}_{}_{}".format(
                    name, aa_descriptor, canvas_descriptor)
                results[description] = img

            for func in (generate_test_006, ):
                points, name = func()
                aggregator = draw_multi_segment_line(canvas, points, antialias)
                img = shade(aggregator, cmap=cmap01)
                description = "{}_{}_{}".format(
                    name, aa_descriptor, canvas_descriptor)
                results[description] = img
    return results

def save_test_images(images):
    """Save all images as  PNG and NetCDF files

    Parameters
    ----------
    images: dict
      A dictionary mapping test case names to xarray images.
    """
    for description, img in images.items():
        save_to_image(img, description)
        save_to_netcdf(img, description)

def load_test_images(images):
    """Load all images from NetCDF files

    Returns
    -------
    loaded: dict
      A dictionary mapping test case names to xarray images.
    """
    loaded = {}
    for description, _ in images.items():
        loaded[description] = load_from_netcdf(description)
    return loaded

def test_antialiasing():
    """Test case for all images.

    Will generate all test cases, then load all test cases from disk and
    compare them with each other

    """
    images  = generate_test_images()
    loaded = load_test_images(images)
    print(list(loaded.keys()))
    for description in images.keys():
        assert (images[description] == loaded[description]).all()

if __name__ == '__main__':
    # Run this to generate the PNG and NetCDF files.
    images  = generate_test_images()
    print(list(images.keys()))
    save_test_images(images)



from __future__ import annotations
import numpy as np
import pandas as pd
import pytest
import datashader as ds
import datashader.transfer_functions as tf


df = pd.DataFrame({'x': np.array(([0.] * 10 + [1] * 10)),
                   'y': np.array(([0.] * 5 + [1] * 5 + [0] * 5 + [1] * 5)),
                   'f64': np.arange(20, dtype='f8')})
df.f64.iloc[2] = np.nan

cvs = ds.Canvas(plot_width=2, plot_height=2, x_range=(0, 1), y_range=(0, 1))
cvs10 = ds.Canvas(plot_width=10, plot_height=10, x_range=(0, 1), y_range=(0, 1))


def test_pipeline():
    pipeline = ds.Pipeline(df, ds.Point('x', 'y'))
    img = pipeline((0, 1), (0, 1), 2, 2)
    agg = cvs.points(df, 'x', 'y', ds.count())
    assert img.equals(tf.shade(agg))

    def color_fn(agg):
        return tf.shade(agg, 'pink', 'red')
    pipeline.color_fn = color_fn
    img = pipeline((0, 1), (0, 1), 2, 2)
    assert img.equals(color_fn(agg))

    def transform_fn(agg):
        return agg + 1
    pipeline.transform_fn = transform_fn
    img = pipeline((0, 1), (0, 1), 2, 2)
    assert img.equals(color_fn(transform_fn(agg)))

    pipeline = ds.Pipeline(df, ds.Point('x', 'y'), ds.sum('f64'))
    img = pipeline((0, 1), (0, 1), 2, 2)
    agg = cvs.points(df, 'x', 'y', ds.sum('f64'))
    assert img.equals(tf.shade(agg))


@pytest.mark.parametrize("line_width", [0.0, 0.5, 1.0, 2.0])
def test_pipeline_antialias(line_width):
    glyph = ds.glyphs.LineAxis0('x', 'y')

    glyph.set_line_width(line_width=line_width)
    assert glyph._line_width == line_width
    assert glyph.antialiased == (line_width > 0)

    pipeline = ds.Pipeline(df, glyph)
    img = pipeline(width=cvs10.plot_width, height=cvs10.plot_height,
                   x_range=cvs10.x_range, y_range=cvs10.y_range)
    agg = cvs10.line(df, 'x', 'y', agg=ds.reductions.count(), line_width=line_width)
    assert img.equals(tf.dynspread(tf.shade(agg)))



"""
Lexer for the datashape grammar.
"""

from __future__ import absolute_import, division, print_function

import re
import ast
import collections

from . import error

# This is updated to include all the token names from _tokens,
# where e.g. _tokens[NAME_LOWER-1] is the entry for NAME_LOWER
__all__ = ['lex', 'Token']

def _str_val(s):
    # Use the Python parser via the ast module to parse the string,
    # since the string_escape and unicode_escape codecs do the wrong thing
    return ast.parse('u' + s).body[0].value.s

# A list of the token names, corresponding regex, and value extraction function
_tokens = [
    ('BOOLEAN',    r'True|False', ast.literal_eval),
    ('NAME_LOWER', r'[a-z][a-zA-Z0-9_]*', lambda x : x),
    ('NAME_UPPER', r'[A-Z][a-zA-Z0-9_]*', lambda x : x),
    ('NAME_OTHER', r'_[a-zA-Z0-9_]*', lambda x : x),
    ('ASTERISK',   r'\*'),
    ('COMMA',      r','),
    ('EQUAL',      r'='),
    ('COLON',      r':'),
    ('LBRACKET',   r'\['),
    ('RBRACKET',   r'\]'),
    ('LBRACE',     r'\{'),
    ('RBRACE',     r'\}'),
    ('LPAREN',     r'\('),
    ('RPAREN',     r'\)'),
    ('ELLIPSIS',   r'\.\.\.'),
    ('RARROW',     r'->'),
    ('QUESTIONMARK', r'\?'),
    ('INTEGER',    r'0(?![0-9])|-?[1-9][0-9]*', int),
    ('STRING', (r"""(?:"(?:[^"\n\r\\]|(?:\\u[0-9a-fA-F]{4})|(?:\\["bfnrt]))*")|""" +
                r"""(?:'(?:[^'\n\r\\]|(?:\\u[0-9a-fA-F]{4})|(?:\\['bfnrt]))*')"""),
                _str_val),
]

# Dynamically add all the token indices to globals() and __all__
__all__.extend(tok[0] for tok in _tokens)
globals().update((tok[0], i) for i, tok in enumerate(_tokens, 1))

# Regex for skipping whitespace and comments
_whitespace = r'(?:\s|(?:#.*$))*'

# Compile the token-matching and whitespace-matching regular expressions
_tokens_re = re.compile('|'.join('(' + tok[1] + ')' for tok in _tokens),
                        re.MULTILINE)
_whitespace_re = re.compile(_whitespace, re.MULTILINE)

Token = collections.namedtuple('Token', 'id, name, span, val')

def lex(ds_str):
    """A generator which lexes a datashape string into a
    sequence of tokens.

    Example
    -------

        import datashape
        s = '   -> ... A... "string" 1234 Blah _eil(# comment'
        print('lexing %r' % s)
        for tok in datashape.lexer.lex(s):
            print(tok.id, tok.name, tok.span, repr(tok.val))
    """
    pos = 0
    # Skip whitespace
    m = _whitespace_re.match(ds_str, pos)
    if m:
        pos = m.end()
    while pos < len(ds_str):
        # Try to match a token
        m = _tokens_re.match(ds_str, pos)
        if m:
            # m.lastindex gives us which group was matched, which
            # is one greater than the index into the _tokens list.
            id = m.lastindex
            tokinfo = _tokens[id - 1]
            name = tokinfo[0]
            span = m.span()
            if len(tokinfo) > 2:
                val = tokinfo[2](ds_str[span[0]:span[1]])
            else:
                val = None
            pos = m.end()
            yield Token(id, name, span, val)
        else:
            raise error.DataShapeSyntaxError(pos, '<nofile>',
                                             ds_str,
                                             'Invalid DataShape token')
        # Skip whitespace
        m = _whitespace_re.match(ds_str, pos)
        if m:
            pos = m.end()
